#include <SPI.h>
#include <TFT_ILI9341_ESP.h>
#include <ESP8266WiFi.h> 
#include "HTTPSRedirect.h"

#define Sampling_Time 5
#define Num_Samples 600
#define Peak_Threshold_Factor 75
#define Minimum_Range 50
#define Minimum_Peak_Separation 50  // 50*5=250 ms
#define DC_Added 10
#define Samples_to_Display 600
#define Display_Sampling 2
#define x_axis_length 300
#define y_axis_length 180
#define Moving_Average_Num 5

#define TFT_GREY 0x5AEB
/* Project : IoT Heart Rate Monitor
 * Written by: Raj Bhatt (02/25/2017)
 * Embedded-Lab.com
 * Uses ESP8266 to compute heart rate using Easy Pulse Plugin sensor and post the data to a Google spreadsheet for remote
 * access.
 */

TFT_ILI9341_ESP tft = TFT_ILI9341_ESP();      

int ADC_Samples[Num_Samples], Index1, Index2, Index3, Peak1, Peak2, Peak3;
long Pulse_Rate, Temp1=1L, PR1, PR2;
int Peak_Magnitude, Peak_Threshold, Minima, Range;

const char* ssid = "Your SSID"; 
const char* password = "Your Password";
const char *GScriptId = "Your GScriptid";

// Push data on this interval 
const int dataPostDelay = 30000;     // 30 sec

const char* host = "script.google.com";
const char* googleRedirHost = "script.googleusercontent.com";

const int httpsPort =     443; 
HTTPSRedirect client(httpsPort);

// Prepare the url (without the varying data)
String url = String("/macros/s/") + GScriptId + "/exec?";

const char* fingerprint = "F0 5C 74 77 3F 6B 25 D7 3B 66 4D 43 2F 7E BC 5B E9 28 86 AD";

void setup() {
  char  ch;  
  tft.init();
  tft.setRotation(1);
  tft.fillScreen(TFT_GREY);
  Serial.begin(115200);

  Serial.println("Connecting to wifi:"); 
  Serial.println(ssid); 
  Serial.flush();
  WiFi.begin(ssid, password); 
  while (WiFi.status() != WL_CONNECTED) { 
            delay(500); 
            Serial.print("."); 
  } 
  Serial.println(" IP address:"); 
  Serial.println(WiFi.localIP());
  Serial.print(String("Connecting to ")); 
  Serial.println(host);
  
  bool flag = false;
  for (int i=0; i<5; i++){
    int retval = client.connect(host, httpsPort);
    if (retval == 1) {
       flag = true;
       break;
    }
    else
      Serial.println("Connection failed. Retrying...");
  }

  // Connection Status, 1 = Connected, 0 is not.
  Serial.println("Connection Status: " + String(client.connected()));
  Serial.flush();
  
  if (!flag){
    Serial.print("Could not connect to server: ");
    Serial.println(host);
    Serial.println("Exiting...");
    Serial.flush();
    return;
  }

  // Data will still be pushed even certification don't match.
  if (client.verify(fingerprint, host)) {
    Serial.println("Certificate match.");
  } else {
    Serial.println("Certificate mis-match");
  }
}  // End setup loop

// This is the main method where data gets pushed to the Google sheet
void postData(String tag, float value){
  if (!client.connected()){
    Serial.println("Connecting to client again..."); 
    client.connect(host, httpsPort);
  }
  String urlFinal = url + "tag=" + tag + "&value=" + String(value);
  client.printRedir(urlFinal, host, googleRedirHost);
}

void loop() {
  Read_ADC_Samples();
  Serial.println("Sample Read Finished ");
  Remove_DC();
  Serial.println("DC component subtracted ");
  Scale_Data();
  Serial.println("Data scaled ");
  if (Range > Minimum_Range){  // ADC range is > 50, otherwise increse gain
    Filter_Data();
    Serial.println("Data Filtered ");
    Compute_Pulse_Rate();
    Serial.println("Pulse rate computed ");
    Display_Samples();
    Serial.println("Sample displayed ");
  
   // Now print heart rate
 
   char strheart[4];
   itoa(Pulse_Rate, strheart, 10);
  }
  postData("HearRate",Pulse_Rate ); 
  
}

void Read_ADC_Samples(){
  for (int i = 0; i < Num_Samples; i++){
     //ADC_Samples[i] = 1023-analogRead(A1);
     ADC_Samples[i] = analogRead(A0);
     delay(5);
     
  }  
}

void Remove_DC(){
  Find_Minima(0);
  Serial.print("Minima = ");
  Serial.println(Minima);
  // Subtract DC (minima) 
  for (int i = 0; i < Num_Samples; i++){
     ADC_Samples[i] = ADC_Samples[i] - Minima;
  }
  Minima = 0;  // New minima is zero  
}

void Scale_Data(){
  // Find peak value
  Find_Peak(0);
  Serial.print("Peak = ");
  Serial.println(Peak_Magnitude);
  Range = Peak_Magnitude - Minima;
  Serial.print("Range = ");
  Serial.println(Range);
  // Sclae from 1 to 1023 
  if(Range > Minimum_Range){
    for (int i = 0; i < Num_Samples; i++){
       ADC_Samples[i] = 1 + ((ADC_Samples[i]-Minima)*1022)/Range;
     
    }
    Find_Peak(0);
    Find_Minima(0);
    Serial.print("Scaled Peak = ");
    Serial.println(Peak_Magnitude);
    Serial.print("Scaled Minima = ");
    Serial.println(Minima);
  }
}


void Filter_Data(){
  int Num_Points = 2*Moving_Average_Num+1;
  for (int i = Moving_Average_Num; i < Num_Samples-Moving_Average_Num; i++){
    int Sum_Points = 0;
    for(int k =0; k < Num_Points; k++){   
      Sum_Points = Sum_Points + ADC_Samples[i-Moving_Average_Num+k]; 
    }    
    ADC_Samples[i] = Sum_Points/Num_Points; 
  } 
}


void Compute_Pulse_Rate(){
  // Detect Peak magnitude and minima
  Find_Peak(Moving_Average_Num);
  Find_Minima(Moving_Average_Num);
  Range = Peak_Magnitude - Minima;
  Peak_Threshold = Peak_Magnitude*Peak_Threshold_Factor;
  Peak_Threshold = Peak_Threshold/100;
 
  // Now detect three peaks 
  Peak1 = 0;
  Peak2 = 0;
  Peak3 = 0;
  Index1 = 0;
  Index2 = 0;
  Index3 = 0;
  // Find first peak
  for (int j = Moving_Average_Num; j < Num_Samples-Moving_Average_Num; j++){
      if(ADC_Samples[j] >= ADC_Samples[j-1] && ADC_Samples[j] > ADC_Samples[j+1] && 
         ADC_Samples[j] > Peak_Threshold && Peak1 == 0){
        Peak1 = ADC_Samples[j];
        Index1 = j; 
      }
      
      // Search for second peak which is at least 10 sample time far
      if(Peak1 > 0 && j > (Index1+Minimum_Peak_Separation) && Peak2 == 0){
         if(ADC_Samples[j] >= ADC_Samples[j-1] && ADC_Samples[j] > ADC_Samples[j+1] && 
         ADC_Samples[j] > Peak_Threshold){
         Peak2 = ADC_Samples[j];
         Index2 = j; 
         } 
      } // Peak1 > 0
      
      // Search for the third peak which is at least 10 sample time far
      if(Peak2 > 0 && j > (Index2+Minimum_Peak_Separation) && Peak3 == 0){
         if(ADC_Samples[j] >= ADC_Samples[j-1] && ADC_Samples[j] > ADC_Samples[j+1] && 
         ADC_Samples[j] > Peak_Threshold){
            Peak3 = ADC_Samples[j];
            Index3 = j; 
         } 
      } // Peak2 > 0
    
  }
  Serial.print("Index1 = ");
  Serial.println(Index1);
  Serial.print("Index2 = ");
  Serial.println(Index2);
  Serial.print("Index3 = ");
  Serial.println(Index3);
  
  
  PR1 = (Index2-Index1)*Sampling_Time; // In milliseconds
  PR2 = (Index3-Index2)*Sampling_Time;
  Serial.print("PR1 = ");
  Serial.println(PR1);
  Serial.print("PR2 = ");
  Serial.println(PR2);
  if(PR1 > 0 && PR2 > 0){
    Pulse_Rate = (PR1+PR2)/2;
    Pulse_Rate = 60000/Pulse_Rate; // In BPM
    Serial.println("Pulse rate");
    Serial.println(Pulse_Rate);
     
  }
}

void Display_Samples(){
 tft.fillScreen(TFT_GREY);
 tft.drawLine(0, 0, 0, y_axis_length, TFT_GREEN);
 tft.drawLine(1, 0, 1, y_axis_length, TFT_GREEN);
 tft.drawLine(0, y_axis_length, x_axis_length, y_axis_length, TFT_GREEN);
 tft.drawLine(0, y_axis_length+1, x_axis_length, y_axis_length+1, TFT_GREEN);

  for(int i = Moving_Average_Num; i < Samples_to_Display; i+=Display_Sampling){
    int y_pos = 1 + ((ADC_Samples[i]-Minima)*(y_axis_length-2))/Range;
    y_pos = y_axis_length-y_pos;
    tft.drawPixel(i/Display_Sampling, y_pos, TFT_WHITE);
    tft.drawPixel(i/Display_Sampling+1, y_pos, TFT_WHITE);
    tft.drawPixel(i/Display_Sampling, y_pos+1, TFT_WHITE);
    tft.drawPixel(i/Display_Sampling+1, y_pos+1, TFT_WHITE);
   // tft.drawPixel(i/Display_Sampling, y_pos-1, TFT_WHITE);
  }
  tft.setTextColor(TFT_YELLOW); 
  tft.setCursor(80, y_axis_length+10,4);
  tft.setTextSize(2);
  tft.print(Pulse_Rate);
  tft.print(" BPM");
}

void Find_Peak(int Num){
  Peak_Magnitude = 0;
  for (int m = Num; m < Num_Samples-Num; m++){
      if(Peak_Magnitude < ADC_Samples[m]){
        Peak_Magnitude = ADC_Samples[m];
     }
  }
}

void Find_Minima(int Num){
  Minima = 1024;
  for (int m = Num; m < Num_Samples-Num; m++){
      if(Minima > ADC_Samples[m]){
         Minima = ADC_Samples[m];
      }
  }
}

